package org.netkernel.layer1.endpoint;

import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;

import org.netkernel.container.IKernel;
import org.netkernel.container.ISpaceListener;
import org.netkernel.layer0.meta.*;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.nkf.impl.NKFResponseReadOnlyImpl;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.*;
import org.netkernel.module.standard.builtin.PrivateFilterEndpoint;
import org.netkernel.module.standard.endpoint.StandardEndpointImpl;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestResponseFieldsImpl;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;
import org.netkernel.urii.impl.Version;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class DynamicImportEndpoint extends StandardEndpointImpl implements ISpaceListener
{
	private ReentrantLock mBindLock = new ReentrantLock();
	private boolean mIsPrivate;
	private ISpaceElements mSpaceElements = SpaceElementsImpl.EMPTY;
	@SuppressWarnings("unchecked")
	private INKFResponseReadOnly[] mConfigResponseDependencies;
	private Document mConfig;
	private Map<String, LogicalEpStruct> mElementMetasById = new LinkedHashMap<String, LogicalEpStruct>();
	private List<ISpaceWithIdentity> mImportedSpaces;
	private boolean mSpaceChanged;

	private static class LogicalEpStruct
	{	public IMetaRepresentation mMeta;
		public ISpace mSpace;

		public LogicalEpStruct(IMetaRepresentation aMeta, ISpace aSpace)
		{	mMeta = aMeta;
			mSpace = aSpace;
		}
	}

	public DynamicImportEndpoint()
	{	this.declareThreadSafe();
		this.declareVerboseUnhandledExceptions(false);
		this.declareUnhandledExceptionsExpired(false);
	}

	public void onCommissionEndpoint(IKernel aKernel, ISpace aSpace) throws Exception
	{	super.onCommissionEndpoint(aKernel, aSpace);
		aKernel.addSpaceListener(this);
		mIsPrivate = (Boolean) getParameter("private");
	}

	public void onDecommissionEndpoint() throws Exception
	{	getKernel().removeSpaceListener(this);
		super.onDecommissionEndpoint();
	}

	public ISpaceElements getParts(INKFRequestContext aContext) throws Exception
	{	boolean locked = mBindLock.isLocked();
		if (locked)
		{ // Prevent recursion if we're doing findImports();
			long expiryTime = getKernel().getTimer().getApproximateTime()
					+ getKernel().getConfiguration().getLong("stdmod.initialisation.hysteresis", 500L);
			TimedExpiryFunction exp = new TimedExpiryFunction("InitialisationHysteresis", expiryTime);
			aContext.getKernelContext().declareDependency(exp, Integer.MAX_VALUE);
			return mSpaceElements;
		}
		bind(aContext);

		return mSpaceElements;
	}

	@Override
	public void onResolve(INKFResolutionContext aContext) throws Exception
	{	if (mBindLock.isLocked())
		{	if (mBindLock.isHeldByCurrentThread())
			{ // Prevent recursion
				return;
			}
			else
			{	long h=getKernel().getConfiguration().getLong("stdmod.initialisation.hysteresis", 500L);
				//long expiryTime = getKernel().getTimer().getApproximateTime() + h;
				boolean aquired = mBindLock.tryLock(h, TimeUnit.MILLISECONDS);
				if (aquired)
				{	mBindLock.unlock();
				}
				else
				{	// we failed to wait for binding to complete - better to be
					// live than have success
					aContext.getKernelContext().declareDependency(new TimedExpiryFunction("Dynamic Import Binding", System.currentTimeMillis()+h), Integer.MAX_VALUE);
					return;
				}
			}
		}
		INKFRequestReadOnly toResolve = aContext.getRequestToResolve();
		bind(aContext);
		
		for (Map.Entry<String, LogicalEpStruct> entry : mElementMetasById.entrySet())
		{	IMetaRepresentation meta = entry.getValue().mMeta;
			if (meta instanceof IEndpointMeta)
			{	if (MetadataUtils.match((IEndpointMeta) meta, toResolve, aContext))
				{	String endpointId = entry.getKey();
					INKFResponse resp = aContext.createResolutionResponse(this, endpointId);
					resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
					break;
				}
			}
		}
	}

	public void onRequest(INKFRequestContext aContext) throws Exception
	{	if (mBindLock.isLocked())
		{	return;
		}

		INKFRequestReadOnly req = aContext.getThisRequest();
		String id;
		if (req.getVerb() == INKFRequestReadOnly.VERB_META)
		{	id = req.getIdentifier();
			LogicalEpStruct leps = mElementMetasById.get(id);
			if (leps != null)
			{	IMetaRepresentation meta = leps.mMeta;
				if (mIsPrivate)
				{	meta = PrivateFilterEndpoint.cloneAsPrivate(meta);
				}
				aContext.createResponseFrom(meta);
			}
		}
		else
		{	id = req.getResolvedElementId();
			LogicalEpStruct struct = mElementMetasById.get(id);
			Utils.delegateRequestIntoDetectingAsync(struct.mSpace, aContext, mConfigResponseDependencies);
		}
	}

	private static final PairList VALIDATION = new PairList(4);
	static
	{	VALIDATION.put("/*/*[name()!='space']", "MSG_DYN_IMP_CONFIG1");
		VALIDATION.put("/*/space/*[name()!='uri' and name()!='version']", "MSG_DYN_IMP_CONFIG2");
		VALIDATION.put("/*/space[version and not(uri)]", "MSG_DYN_IMP_CONFIG3");
	}

	private void bind(INKFRequestContext aContext) throws Exception
	{
		boolean spaceChanged=mSpaceChanged;
		boolean bind = mConfigResponseDependencies == null || spaceChanged;
		if (!bind)
		{	for (int i = 0; i < mConfigResponseDependencies.length; i++)
			{	if (mConfigResponseDependencies[i].isExpired())
				{	bind = true;
					break;
				}
			}
		}
		mSpaceChanged=false;
		
		if (bind)
		{	mBindLock.lock();
			try
			{	INKFResponseReadOnly configResponse;
				if (mConfigResponseDependencies == null || mConfigResponseDependencies[0].isExpired())
				{ // only reload config if it has changed
					Object[] repResp = Utils.sourceResponseAsDOMAvoidingTransrept(aContext, "param:config");
					configResponse = (NKFResponseReadOnlyImpl) repResp[1];
					mConfig = (Document) repResp[0];

					PairList validation = FastSchematron.validate(mConfig.getDocumentElement(), VALIDATION);
					if (validation.size() > 0)
					{
						String m = FastSchematron.formatValidationError(validation, aContext);
						throw aContext.createFormattedException("EX_DYN_IMP_CONFIG", "MSG_RAW", null, null, m);
					}
				}
				else
				{	configResponse = mConfigResponseDependencies[0];
				}

				mImportedSpaces = new ArrayList<ISpaceWithIdentity>();
				for (Element se = XMLUtils.getFirstChildElement(mConfig.getDocumentElement()); se != null; se = XMLUtils.getNextSiblingElement(se))
				{	String configURI = XMLUtils.getText(XMLUtils.getChildElementNamed(se, "uri")).trim();
					Version v = null;
					Element versionElement = XMLUtils.getChildElementNamed(se, "version");
					if (versionElement != null)
					{	v = new Version(XMLUtils.getText(versionElement).trim());
					}
					ISpaceWithIdentity space = getKernel().getSpace(new SimpleIdentifierImpl(configURI), v, v);
					if (space == null)
					{	aContext.logFormatted(INKFRequestContext.LEVEL_WARNING, "LOG_DYNAMIC_IMPORT_FAIL", configURI, v, getSpace());
					}
					else
					{	mImportedSpaces.add(space);
					}
				}

				LinkedHashMap elementMetasById = new LinkedHashMap<String, LogicalEpStruct>();
				TupleList pl = new TupleList(3, 8);
				List<INKFResponseReadOnly> metaResponses = new ArrayList<INKFResponseReadOnly>();
				for (ISpace s : mImportedSpaces)
				{	INKFResponseReadOnly<ISpaceMeta> smResp = getSpaceMeta(s, aContext);
					metaResponses.add(smResp);
					ISpaceMeta sm = smResp.getRepresentation();
					if (sm != null)
					{	ISpaceElements imported = sm.getElements();
						for (int j = 0; j < imported.size(); j++)
						{	IIdentifier elementId = imported.getIdentifier(j);
							if (elementId != null)
							{	pl.put(elementId, s, this);
								String elementIdString = elementId.toString();
								IMetaRepresentation meta = getElementMeta(elementIdString, s, aContext);
								LogicalEpStruct leps = new LogicalEpStruct(meta, s);
								LogicalEpStruct removed=(LogicalEpStruct)elementMetasById.put(elementIdString, leps);
								if (removed!=null)
								{	elementMetasById.put(elementIdString, removed);
								}
							}
						}
					}
				}
				mSpaceElements = new SpaceElementsImpl(pl);
				mElementMetasById = elementMetasById;
				INKFResponseReadOnly[] configResponseDependencies= new INKFResponseReadOnly[metaResponses.size() + 1];
				configResponseDependencies[0] = configResponse;
				for (int i = 0; i < metaResponses.size(); i++)
				{	configResponseDependencies[i + 1] = metaResponses.get(i);
				}
				mConfigResponseDependencies = configResponseDependencies;
			}
			catch (NKFException e)
			{ // failed to get config
				//aContext.logFormatted(INKFLocale.LEVEL_WARNING, "LOG_DYNAMIC_IMPORT_CONFIG", getSpace(),e.getDeepestId()+" "+e.getDeepestMessage());
			}
			finally
			{	mBindLock.unlock();
			}
		}
		else
		{	for (int i = 0; i < mConfigResponseDependencies.length; i++)
			{	aContext.getKernelContext().declareDependency(mConfigResponseDependencies[i]);
			}
		}
	}

	private IMetaRepresentation getElementMeta(String aElementId, ISpace aSpace, INKFRequestContext aContext) throws NKFException
	{	INKFRequest req = aContext.createRequest(aElementId);
		req.setVerb(INKFRequestReadOnly.VERB_META);
		IRequestScopeLevel rctx = RequestScopeLevelImpl.createOrphanedRootScopeLevel(aSpace, aContext.getKernelContext().getRequestScope());
		req.setRequestScope(rctx);
		req.setRepresentationClass(IMetaRepresentation.class);
		return (IMetaRepresentation) aContext.issueRequest(req);
	}

	private INKFResponseReadOnly<ISpaceMeta> getSpaceMeta(ISpace delegateSpace, INKFRequestContext aContext) throws NKFException
	{	INKFRequest req = aContext.createRequest("");
		req.setVerb(INKFRequestReadOnly.VERB_META);
		IRequestScopeLevel rctx = RequestScopeLevelImpl.createOrphanedRootScopeLevel(delegateSpace, aContext.getKernelContext().getRequestScope());
		req.setRequestScope(rctx);
		req.setRepresentationClass(ISpaceMeta.class);
		return aContext.issueRequestForResponse(req);
	}

	public void spaceChanged()
	{	mSpaceChanged = true;
	}

	public IRequestResponseFields getState()
	{	RequestResponseFieldsImpl fields = new RequestResponseFieldsImpl(super.getState());
		HDSBuilder b = new HDSBuilder();
		b.pushNode("imports");
		if (mImportedSpaces != null)
		{	for (ISpaceWithIdentity space : mImportedSpaces)
			{	b.pushNode("import");
				b.addNode("uri", space.getIdentifier().toString());
				b.addNode("version", space.getVersion().toString());
				b.popNode();
			}
			fields.put("imports", b.getRoot());
		}
		return fields;
	}
	
}