package org.netkernel.layer1.endpoint;

import org.netkernel.layer0.meta.impl.IdentifierArgumentMetaImpl;
import org.netkernel.layer0.nkf.INKFExpiryFunction;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.urii.ParsedIdentifierImpl;
import org.netkernel.layer0.util.GoldenThreadExpiryFunction;
import org.netkernel.layer1.util.GoldenThreadGroup;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.impl.RequestResponseFieldsImpl;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.ReentrantLock;

public class GoldenThreadAccessorWithRecalc extends StandardAccessorImpl
{
	private static final String ARG_ID="id";
	private static final String ARG_RECALC="recalcRequest";
	private static GoldenThreadAccessorWithRecalc sSingleton;
	
	private GoldenThreadGroup mGroup;
	private GoldenThreadExpiryFunction mGroupDirtyExpiry;
	private ReentrantLock mSyncLock=new ReentrantLock();
	
	private ConcurrentHashMap<Object, Map<String, INKFRequest>> RecalculateRequestMap=new ConcurrentHashMap();
	
	protected static GoldenThreadAccessorWithRecalc getSingleton()
	{	return sSingleton;
	}
	
	protected GoldenThreadGroup getGroup()
	{	return mGroup;
	}
	
	public GoldenThreadExpiryFunction getGroupDirtyExpiry()
	{	mSyncLock.lock();
		try
		{	if (mGroupDirtyExpiry==null)
			{	mGroupDirtyExpiry=new GoldenThreadExpiryFunction("GoldenThreadGroupDirtyExpiry");
			}
			return mGroupDirtyExpiry;
		}
		finally
		{	mSyncLock.unlock();
		}
	}
	
	protected void markDirty()
	{	mSyncLock.lock();
		try
		{	if (mGroupDirtyExpiry!=null)
			{	mGroupDirtyExpiry.invalidate();
				mGroupDirtyExpiry=null;
			}
		}
		finally
		{	mSyncLock.unlock();
		}
	}
	
	public GoldenThreadAccessorWithRecalc()
	{	declareThreadSafe();
		mGroup = new GoldenThreadGroup();
		declareArgument(new IdentifierArgumentMetaImpl("id", null, null));
		sSingleton=this;
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{	String activeType=aContext.getThisRequest().getArgumentValue(ParsedIdentifierImpl.ARG_ACTIVE_TYPE);
		if (activeType.equals("attachRecalcGoldenThread"))
		{	onAttach(aContext);
		}
		else if (activeType.equals("cutRecalcGoldenThread"))
		{	onCut(aContext);
		}
		else if (activeType.equals("cutRecalcGoldenThreadAndRecalc"))
		{	onCutAndRecalc(aContext);
		}
	}
	
	@Override
	public void onExists(INKFRequestContext aContext) throws Exception
	{		String activeType=aContext.getThisRequest().getArgumentValue(ParsedIdentifierImpl.ARG_ACTIVE_TYPE);
			if (activeType.equals("attachRecalcGoldenThread"))
			{	
				String id=aContext.getThisRequest().getArgumentValue(ARG_ID);
				INKFExpiryFunction gt=mGroup.exists(id);
				INKFResponse response=aContext.createResponseFrom(gt!=null);
				response.setExpiry(INKFResponse.EXPIRY_MIN_FUNCTION_DEPENDENT, gt);
			}
			else
			{	throw new Exception("EXISTS not supported on active:cutRecalcGoldenThread");				
			}
	}

	public void onAttach(INKFRequestContext aContext) throws Exception
	{	
		mSyncLock.lock();
		try
		{
			INKFResponse response=aContext.createResponseFrom(null);
			response.setNoCache();
			
			INKFRequestReadOnly thisReq=aContext.getThisRequest();
			List<INKFExpiryFunction> threads = new ArrayList();
			INKFRequest recalcRequest=null;
			ArrayList<String> ids=new ArrayList();
			for (int i=0; i<thisReq.getArgumentCount(); i++)
			{	String argName=thisReq.getArgumentName(i);
				if (argName.equals(ARG_ID))
				{	String threadId=thisReq.getArgumentValue(i);
					INKFExpiryFunction goldenThread=mGroup.attach(threadId);
					threads.add(goldenThread);
					ids.add(threadId);
				}
			}
			for (int i=0; i<thisReq.getArgumentCount(); i++)
			{	String argName=thisReq.getArgumentName(i);
				if (argName.equals(ARG_RECALC))
				{	String recalcRequestId = thisReq.getArgumentValue(i);
					recalcRequest=aContext.source(recalcRequestId, INKFRequest.class);
					for(String threadId : ids)
					{	Map<String, INKFRequest> recalcRequestMap = null;
						recalcRequestMap=RecalculateRequestMap.get(threadId);
						if(recalcRequestMap==null)
						{	recalcRequestMap=new ConcurrentHashMap<String, INKFRequest>();						
						}
						recalcRequestMap.put(recalcRequest.getIdentifier(), recalcRequest);
						RecalculateRequestMap.put(threadId, recalcRequestMap);
					}
				}
			}
			if (threads.size()==1)
			{	response.setExpiry(INKFResponse.EXPIRY_MIN_FUNCTION_DEPENDENT, threads.get(0));
			}
			else
			{	response.setExpiry(INKFResponse.EXPIRY_MIN_FUNCTION_DEPENDENT, new ExpiryUnion(threads));
			}
			markDirty();
		}
		finally
		{	mSyncLock.unlock();
		}
	}

	public void onCut(INKFRequestContext aContext) throws Exception
	{	INKFRequestReadOnly thisReq=aContext.getThisRequest();
		for (int i=0; i<thisReq.getArgumentCount(); i++)
		{	String argName=thisReq.getArgumentName(i);
			if (argName.equals(ARG_ID))
			{	String threadId = thisReq.getArgumentValue(i);
				mGroup.cut(threadId);
				RecalculateRequestMap.remove(threadId);
			}
		}
		INKFResponse resp=aContext.createResponseFrom(null);
		resp.setNoCache();
		markDirty();
	}
	
	public void onCutAndRecalc(INKFRequestContext aContext) throws Exception
	{	mSyncLock.lock();
		try
		{
			INKFRequestReadOnly thisReq=aContext.getThisRequest();
			ArrayList<Map<String, INKFRequest>> recalcRequestList=new ArrayList();
			for (int i=0; i<thisReq.getArgumentCount(); i++)
			{	String argName=thisReq.getArgumentName(i);
				if (argName.equals(ARG_ID))
				{	String threadId = thisReq.getArgumentValue(i);
					mGroup.cut(threadId);
					Map<String, INKFRequest> recalcMap=RecalculateRequestMap.remove(threadId);
					if(recalcMap!=null)
					{	recalcRequestList.add(recalcMap);
					}
				}
			}
			INKFResponse resp=aContext.createResponseFrom(null);
			resp.setNoCache();
			markDirty();
			//Issue Recalculation Requests
			for(Map<String, INKFRequest> recalcMap : recalcRequestList)
			{	for(INKFRequest req : recalcMap.values())
				{	aContext.issueAsyncRequest(req);
				}
			}
		}
		finally
		{	mSyncLock.unlock();
		}
	}
	
	public IRequestResponseFields getState()
	{	RequestResponseFieldsImpl fields=new RequestResponseFieldsImpl(super.getState());
		fields.put("size", mGroup.size());
		return fields;
	}
	
	private static class ExpiryUnion implements INKFExpiryFunction
	{
		private INKFExpiryFunction[] mElements;
		
		public ExpiryUnion(List<INKFExpiryFunction> aElements)
		{	mElements=new INKFExpiryFunction[aElements.size()];
			mElements=aElements.toArray(mElements);
		}

		public boolean isExpired(long aNow)
		{	for (INKFExpiryFunction expiry: mElements)
			{	if (expiry.isExpired(aNow))
				{	return true;
				}
			}
			return false;
		}
		
		public String toString()
		{	StringBuilder sb = new StringBuilder(64);
			for (int i=0; i<mElements.length; i++)
			{	if (i!=0)
				{	sb.append(", ");
				}
				sb.append(mElements[i].toString());
			}
			return sb.toString();
		}
		
	}
}