package org.netkernel.layer1.endpoint;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.ReentrantLock;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.netkernel.layer0.meta.impl.SourcedArgumentMetaImpl;
import org.netkernel.layer0.nkf.INKFAsyncRequestListener;
import org.netkernel.layer0.nkf.INKFLocale;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.urii.ParsedIdentifierImpl;
import org.netkernel.layer0.util.RequestBuilder;
import org.netkernel.layer0.util.RequestBuilder.Arguments;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.IResolution;
import org.netkernel.urii.ISpace;
import org.netkernel.util.Utils;
import org.w3c.dom.Document;

public class GoldenThreadCutListenerAccessor extends StandardAccessorImpl
{
	private ConcurrentHashMap<CutListener,CutListener> mListeners=new ConcurrentHashMap<>();
	private static GoldenThreadCutListenerAccessor sInstance;
	
	public static GoldenThreadCutListenerAccessor getInstance()
	{	return sInstance;
	}
	
	public GoldenThreadCutListenerAccessor()
	{	this.declareThreadSafe(false);
		this.declareArgument(new SourcedArgumentMetaImpl("id", "Golden thread identifier regular expression", null, new Class[]{String.class}));
		this.declareArgument(new SourcedArgumentMetaImpl("request", "declarative request or INKFRequest to issue when thread is cut", null, new Class[]{INKFRequest.class, Document.class}));
		sInstance=this;
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		String type=aContext.getThisRequest().getArgumentValue(ParsedIdentifierImpl.ARG_ACTIVE_TYPE);
		if (type.equals("addGoldenThreadCutListener"))
		{	onAdd(aContext);
		}
		else if (type.equals("removeGoldenThreadCutListener"))
		{	onRemove(aContext);
		}
	}
	
	public void onAdd(INKFRequestContext aContext) throws Exception
	{	CutListener cl=new CutListener(aContext);
		boolean added=mListeners.putIfAbsent(cl,cl)==null;
		INKFResponse resp=aContext.createResponseFrom(added);
		resp.setNoCache();
	}
	
	
	public void onRemove(INKFRequestContext aContext) throws Exception
	{	CutListener cl=new CutListener(aContext);
		boolean removed=(mListeners.remove(cl)!=null);
		INKFResponse resp=aContext.createResponseFrom(removed);
		resp.setNoCache();
	}
	
	public void onCut(INKFRequestContext aContext, String aThreadId)
	{
		for (CutListener cl : mListeners.keySet())
		{	if (cl.matches(aThreadId))
			{	try
				{	cl.issueCutNotificationRequest(aContext, aThreadId);
				}
				catch (Exception e)
				{	aContext.logRaw(INKFLocale.LEVEL_WARNING, Utils.throwableToString(e));
				}
			}
		}
	}
	
	private static class CutListener
	{
		private final RequestBuilder mRequestBuilder;
		private final INKFRequest mRequest;
		private final ISpace mIssuingSpace;
		private final int mHashCode;
		private final Matcher mMatcher;
		private final ReentrantLock mSyncLock=new ReentrantLock();
		
		public CutListener(INKFRequestContext aContext) throws Exception
		{
			Object requestArg=aContext.source("arg:request",Object.class);
			String requestString;
			if (requestArg instanceof INKFRequest)
			{	INKFRequest req=(INKFRequest)requestArg;
				IResolution resolution=aContext.getKernelContext().resolveRequest(req);
				mIssuingSpace=resolution.getScope().getSpace();
				requestString=req.getIdentifier();
				mRequest=req;
				mRequestBuilder=null;
			}
			else
			{	Document declRequestNode=aContext.transrept(requestArg,Document.class);
				mRequestBuilder=new RequestBuilder(declRequestNode, aContext.getKernelContext().getKernel().getLogger());
				Arguments args=new Arguments();
				args.addArgumentByValue("id", "dummy");
				INKFRequest req=mRequestBuilder.buildRequest(aContext, args, null);
				IResolution resolution=aContext.getKernelContext().resolveRequest(req);
				mIssuingSpace=resolution.getScope().getSpace();
				requestString=XMLUtils.toXML(declRequestNode, false, true);
				mRequest=null;
			}
			
			String regex=aContext.source("arg:id",String.class);
			mMatcher=Pattern.compile(regex).matcher("");
			
			mHashCode=requestString.hashCode()^mIssuingSpace.hashCode()^regex.hashCode();
			
		}
		
		@Override
		public boolean equals(Object aOther)
		{	boolean match=false;
			if (aOther instanceof CutListener)
			{	CutListener other=(CutListener)aOther;
				match=mHashCode==other.mHashCode && mIssuingSpace==other.mIssuingSpace && mMatcher.pattern().pattern().equals(other.mMatcher.pattern().pattern());
			}
			return match;
		}
		
		@Override
		public int hashCode()
		{	return mHashCode;
		}
		
		public void issueCutNotificationRequest(final INKFRequestContext aContext, String aThreadId) throws Exception
		{
			INKFRequest req;
			if (mRequestBuilder!=null)
			{	Arguments args=new Arguments();
				args.addArgument("arg:id", aThreadId);
				req=mRequestBuilder.buildRequest(aContext, args, null);
				IRequestScopeLevel scope=RequestScopeLevelImpl.createOrphanedRootScopeLevel(mIssuingSpace, aContext.getKernelContext().getRequestScope());
				req.setRequestScope(scope);
			}
			else
			{	req=mRequest;
			}
			
			
			aContext.issueAsyncRequest(req).setListener(new INKFAsyncRequestListener()
			{
				public void receiveResponse(INKFResponseReadOnly aResponse, INKFRequestContext aContext) throws Exception
				{	aContext.setNoResponse();
				}
				
				@Override
				public void receiveException(NKFException aException, INKFRequest aRequest, INKFRequestContext aContext) throws Exception
				{	aContext.logRaw(INKFLocale.LEVEL_WARNING, Utils.throwableToString(aException));
					aContext.setNoResponse();
				}
			});
		}
		
		public boolean matches(Object aThreadId)
		{	mSyncLock.lock();
			try
			{	return mMatcher.reset(aThreadId.toString()).matches();
			}
			finally
			{	mSyncLock.unlock();
			}
		}
		
	}
}
