package org.netkernel.layer1.endpoint;

import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.layer0.bnf.BNFGrammarFactory;
import org.netkernel.layer0.bnf.GrammarParseException;
import org.netkernel.layer0.bnf.IGrammarGroup;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.impl.EndpointArgumentMetaImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.util.GoldenThreadExpiryFunction;
import org.netkernel.module.standard.StandardSpace;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.module.standard.endpoint.TransparentOverlayImpl;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.ClassLoaderWithExports;
import org.netkernel.urii.ISpace;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

public class ScratchPadOverlay extends TransparentOverlayImpl
{
	private ScratchPadProtoSpace mSpace;
	
	public ScratchPadOverlay()
	{	declareThreadSafe();
		declareUnhandledExceptionsExpired(false);
	}
	
	public void onCommissionEndpoint(IKernel kernel, ISpace aSpace) throws Exception
	{	super.onCommissionEndpoint(kernel, aSpace);
		mSpace= new ScratchPadProtoSpace((StandardSpace)aSpace, this, kernel);
		mSpace.onCommissionSpace(kernel);
	}	
	
	public void onRequest(String aID, INKFRequestContext aContext) throws Exception
	{	
		INKFRequest reqOut=aContext.getThisRequest().getIssuableClone();
		IRequestScopeLevel scope=aContext.getKernelContext().getThisKernelRequest().getRequestScope();
		ScratchPadSpace space=new ScratchPadSpace(getStandardSpace(),mSpace);
		space.onCommissionSpace(aContext.getKernelContext().getKernel());
		scope=RequestScopeLevelImpl.createDurableScopeLevel(space, scope);
		scope=RequestScopeLevelImpl.createNonDurableScopeLevel(getDelegateSpace(), scope);
		reqOut.setRequestScope(scope);
		INKFResponseReadOnly resp=aContext.issueRequestForResponse(reqOut);
		aContext.createResponseFrom(resp);	
	}
	
	public ISpace[] getSpaces(INKFRequestContext aContext) throws Exception
	{	ISpace[] result = super.getSpaces(aContext);
		ISpace[] result2 = new ISpace[result.length+1];
		System.arraycopy(result, 0, result2, 1, result.length);
		result2[0]=mSpace;
		return result2;
	}		
	
	private static class ScratchData
	{	public Object mRepresentation;
		public GoldenThreadExpiryFunction mExpiry;
		public static Object NULL="ScratchDataNull";
		public ScratchData(Object aRepresentation)
		{	mRepresentation=aRepresentation;
			mExpiry=new GoldenThreadExpiryFunction("ScratchPad");
		}
	}
	
	private static class ScratchPadSpace extends StandardSpace
	{
		private Map<String,ScratchData> mValueMap = new HashMap<String,ScratchData>();
		

		public ScratchPadSpace(StandardSpace aOwningSpace, ScratchPadProtoSpace aProtoSpace)
		{	super(aOwningSpace, aOwningSpace.getOwningModule(),null);
			setMappings(aProtoSpace.getEndpoints());
			setName("HTTP Session Space");
			setState(ISpaceMeta.STATE_PROTOSPACE, aProtoSpace );		
		}
		
		@Override
		public ClassLoaderWithExports getClassLoader()
		{	return null;
		}

		public synchronized void handleRequestFromEndpoint(INKFRequestContext helper) throws Exception
		{	String identifier=helper.getThisRequest().getIdentifier();
			ScratchData data=null;
			switch(helper.getThisRequest().getVerb())
			{	case INKFRequestReadOnly.VERB_SOURCE:
					data=mValueMap.get(identifier);
					if (data==null || data.mRepresentation==ScratchData.NULL)
					{	throw new Exception(identifier+" not found");
					}
					helper.createResponseFrom(data.mRepresentation).setExpiry(INKFResponse.EXPIRY_FUNCTION, data.mExpiry);
				break;
				case INKFRequestReadOnly.VERB_SINK:
					Object primary = helper.getThisRequest().getPrimary();
					data=mValueMap.get(identifier);
					if (data!=null)
					{	data.mExpiry.invalidate();
					}
					data=new ScratchData(primary);
					mValueMap.put(identifier,data);
				break;
				case INKFRequestReadOnly.VERB_DELETE:
					data=mValueMap.remove(identifier);
					if (data!=null)
					{	data.mExpiry.invalidate();
					}
					helper.createResponseFrom(data!=null);
				break;
				case INKFRequestReadOnly.VERB_EXISTS:
					boolean exists=false;
					data= mValueMap.get(identifier);
					if (data==null)
					{	data=new ScratchData(ScratchData.NULL);
						mValueMap.put(identifier,data);
					}
					else
					{	exists=data.mRepresentation!=ScratchData.NULL;
					}
					helper.createResponseFrom(exists).setExpiry(INKFResponse.EXPIRY_FUNCTION, data.mExpiry);
				break;
			}
		}
		
		public String toString()
		{	return "ScratchPadSpace";
		}
	}
	
	private static class ScratchPadProtoSpace extends StandardSpace
	{
		private final IStandardEndpoint[] mEndpoints;
		
		public ScratchPadProtoSpace(StandardSpace aOwningSpace, ScratchPadOverlay aOverlay, IKernel aKernel) throws Exception
		{	super(aOwningSpace,aOwningSpace.getOwningModule(),null);
			IStandardEndpoint ep1 = new ScratchPadEndpoint(aKernel);
			mEndpoints = new IStandardEndpoint[]{ep1};	
			setMappings(mEndpoints);
			setName("Scratchpad Space");
		}
		
		public IStandardEndpoint[] getEndpoints()
		{	return mEndpoints;
		}
	}
	
	private static class ScratchPadEndpoint extends StandardAccessorImpl
	{
		public ScratchPadEndpoint(IKernel aKernel) throws NKFException, GrammarParseException
		{	declareThreadSafe();
			declareInhibitCheckForBadExpirationOnMutableResource();
			declareName("ScratchPadEndpoint");
			IGrammarGroup g1=BNFGrammarFactory.createGroup(null);
			BNFGrammarFactory.createText(g1, "scratch:");
			IGrammarGroup g2=BNFGrammarFactory.createGroup(g1,"path",null);
			BNFGrammarFactory.createRegex(g2, BNFGrammarFactory.BUILTIN_ANYTHING);			
			ILogger logger=aKernel.getLogger();
			IIdentifierGrammar g=BNFGrammarFactory.createGrammar(g1,logger);
			declareGrammar(g);
			declareArgument(new EndpointArgumentMetaImpl("path", "a unique resource path for storing state", null, null, null, false, null));
			setParameterRaw("id", "layer1:scratch:scheme", aKernel);		
		}
		
		public void onSource(INKFRequestContext aContext) throws Exception
		{	innerOnRequest(aContext);
		}
		public void onSink(INKFRequestContext aContext) throws Exception
		{	innerOnRequest(aContext);
		}
		public void onExists(INKFRequestContext aContext) throws Exception
		{	innerOnRequest(aContext);
		}
		public void onDelete(INKFRequestContext aContext) throws Exception
		{	innerOnRequest(aContext);
		}
		private void innerOnRequest(INKFRequestContext aContext) throws Exception
		{	ScratchPadSpace space=(ScratchPadSpace)aContext.getKernelContext().getThisKernelRequest().getRequestScope().getSpace();
			space.handleRequestFromEndpoint(aContext);
		}
	}
}
